#!/usr/bin/python3

from __future__ import annotations
from typing import List, Tuple
from re import compile as reg_compile, Pattern
from functools import reduce
from datetime import datetime

from .message import Message, MessageIndex
from .user import User


class Chat(object):
    '''
        A whole chat, which is generally exported into a text file ( *.txt )
        from WhatsApp, is read and processed & converted into Chat object.

        *Now holding easily retrievable information related to chat time span,
        can be useful in setting an informative title in plots*
    '''

    def __init__(self, users: List[User]):
        self.users = users
        self._messageCount = 0
        # this two can be used in plot titles to denote
        # chat expansion period in better way
        self._startDate = None
        self._endDate = None

    @property
    def messageCount(self) -> int:
        '''
            Returns calculated # of messages present in this chat
        '''
        return self._messageCount

    @messageCount.setter
    def messageCount(self, v):
        self._messageCount = v

    @property
    def startDate(self) -> datetime:
        try:
            return datetime.strptime(self._startDate, r'%d/%m/%y, %I:%M %p')
        except ValueError:
            return datetime.strptime(self._startDate, r'%d/%m/%Y, %I:%M %p')
        except Exception:
            return self._startDate

    @startDate.setter
    def startDate(self, dt):
        self._startDate = dt

    @property
    def endDate(self) -> datetime:
        try:
            return datetime.strptime(self._endDate, r'%d/%m/%y, %I:%M %p')
        except ValueError:
            return datetime.strptime(self._endDate, r'%d/%m/%Y, %I:%M %p')
        except Exception:
            return self._endDate

    @endDate.setter
    def endDate(self, dt):
        self._endDate = dt

    def getUser(self, name: str) -> User:
        '''
            Finds out whether a certain User participated in this chat or not
            ( by its name )

            In case of failure, returns None
        '''
        return reduce(lambda acc, cur: cur if cur.name ==
                      name else acc, self.users, None)

    def getUserByMessageId(self, idx: int) -> User:
        '''
            Helps in finding which user sent a certain message,
            by its index

            In case of invalid message index, returns None
        '''
        return None if idx < 0 or idx >= self.messageCount else reduce(lambda acc, cur:
                                                                       cur if reduce(
                                                                           lambda accInner, curInner: True if curInner.index == idx else accInner,
                                                                           cur.messages, False) else acc,
                                                                       self.users, None)

    @staticmethod
    def importFromText(filePath: str) -> Chat:
        '''
            Given text file path, returns chat object, 
            containing whole chat, which can be 
            easily played around with 
        '''

        def _getRegex() -> Pattern:
            '''
                Regex to be used for extracting timestamp of
                a certain message from `*.txt` file
            '''
            return reg_compile(
                r'(\d{1,2}/\d{1,2}/\d{2,4}, \d{1,2}\:\d{1,2} [a|p]m)')

        def _splitByDate(pattern: Pattern, content: str) -> List[str]:
            '''
                splitting whole *.txt file content using
                date extraction regex we just built
            '''
            return pattern.split(content)[1:]

        def _groupify(splitted: List[str]) -> List[Tuple[str, str]]:
            '''
                In previous closure we splitted whole text file content by date,
                which needs to be fixed by pairing messages with its corresponding timestamp
            '''
            grouped = []
            for i in range(0, len(splitted), 2):
                grouped.append((splitted[i], splitted[i+1]))

            # setting start and end time of chat under lens
            obj.startDate = grouped[0][0]
            obj.endDate = grouped[-1][0]

            return grouped

        def _getUser(text: str) -> str:
            '''
                Helps in extracting participating user name from message ( text )
            '''
            matchObj = reg_compile(r'(?<=\s-\s).+?(?=:)').search(text)
            return matchObj.group() if matchObj else ''

        def _getMessage(text: str) -> str:
            '''
                Extracts actual message sent by a certain user using regex
            '''
            return reg_compile(r'\s-\s.+?(?=:):\s*').sub('', text)

        def _createUserObject(acc: List[User], content: Tuple[str, str]) -> List[User]:
            '''
                Construction of User objects ( who participated in chat ) along with
                messages sent by them, done in this closure
            '''
            # if we can't extract username from message text, it's not a message of this chat, so we just ignore it
            userName = _getUser(content[1])
            if not userName:
                return acc  # returning in unchanged form
            msg = Message(msgIndex.index, _getMessage(
                content[1]), content[0])

            # updating count of message in chat
            obj.messageCount = obj.messageCount + 1

            # incrementing index for next message, which lets us sequentially ordering messages as they appeared in chat
            msgIndex.increment()
            found = reduce(lambda accInner, curInner: curInner if curInner.name ==
                           userName else accInner, acc, None)
            if not found:
                acc.append(User(userName, [msg]))
            else:
                found.messages.append(msg)
            return acc

        obj = Chat([])
        msgIndex = MessageIndex(0)

        with open(filePath, mode='r', encoding='utf-8', errors='ignore') as fd:
            obj.users = reduce(
                _createUserObject, _groupify(
                    _splitByDate(_getRegex(), fd.read())), [])

        return obj


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations

'''
    Holds record of elapsed time between two messages sent in a Chat,
    where messages are identified using their unique `message id`

    Elapsed time will be in second ( unit )
'''


class DifferenceBetweenMessages(object):
    def __init__(self, idx1: int, idx2: int, diff: int):
        self.msgOne = idx1
        self.msgTwo = idx2
        self.elapsedTime = diff

    def __eq__(self, val: DifferenceBetweenMessages) -> bool:
        super().__eq__(val)
        return self.elapsedTime == val.elapsedTime

    def __str__(self) -> str:
        super().__str__()
        return 'Elapsed time b/w msg {} & {} -- {} s'.format(self.msgOne, self.msgTwo, self.elapsedTime)


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from functools import reduce
from collections import Counter
from math import floor

from .model.chat import Chat

def analyze(chat: Chat):
    return removeSpecialCharacters(Counter(
        reduce(lambda acc, cur:
               acc + reduce(lambda accInner,
                            curInner: accInner + curInner.content.split(), cur.messages, []),
               chat.users, [])))


def isASCIISpecialCharacter(c: str) -> bool:
    # here I'm trying to determine whether this is an ASCII special char or not
    return True if len(c) == 1 and (0 <= ord(c) <= 255) and not (96 < ord(c) < 123) and not (64 < ord(c) < 91) else False


def getMedian(frm: int, to: int) -> int:
    return floor((frm + to)/2)


def removeSpecialCharacters(words: Counter) -> Counter:
    # removing all those keys, which are having any combination of only ASCII special characters
    for i in list(words.keys()):
        if all(map(isASCIISpecialCharacter, i)):
            words.pop(i)
    medianV = getMedian(min(words.values()), max(words.values()))
    # removing all those keys from dictionary, which are having value lesser than median
    for i in list(words.keys()):
        if words[i] < medianV:
            words.pop(i)
    return words


if __name__ == '__main__':
    print('[!]This module is designed to be working as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from typing import List, Tuple, Dict
from functools import reduce
from sys import argv
from os.path import join, exists
from os import getenv
from matplotlib import pyplot as plt
from time import time
from multiprocessing import cpu_count
import ray

from .util import (
    plotContributionInChatByUser,
    plotContributionOfUserByHour,
    shadeContactName,
    plotActivityOfUserByMinute,
    mergeMessagesFromUsersIntoSequence,
    classifyMessagesOfChatByDate,
    plotActivenessOfChatByDate,
    directoryBuilder,
    getConversationInitializers,
    plotConversationInitializerStat,
    plotActivityHeatMap
)
from .model.chat import Chat
from .emoji_data.get import (
    exportToFile,
    importFromFile
)
from .emoji import (
    findNonASCIICharactersinText,
    findEmojisInText,
    findEmojiUsage,
    plotEmojiUsage
)

# initializing ray - powering concurrent processing of plotting ops !!!
ray.init(num_cpus=cpu_count())


def _getSupportedOutputFormats() -> Dict[str, str]:
    '''
        Tries to determine supported backends in host machine,
        only these formats can be used for exporting a plot
    '''
    fig = plt.figure()

    formats = fig.canvas.get_supported_filetypes()

    plt.close(fig)
    return formats


def _makeOutputChoice() -> str:
    options = _getSupportedOutputFormats()
    if not options:
        return None

    for i, (k, v) in enumerate(options.items()):
        print('{} ) {} [ .{} ]'.format(i+1, v, k))

    try:
        choice = int(input('\n[?]Choose desired output format : '))
        if choice not in range(1, len(options) + 1):
            raise Exception('Bad Input')
        return list(options.keys())[choice-1]
    except EOFError:
        return None
    except Exception:
        return None


def _calculatePercentageOfSuccess(stat: List[bool]) -> float:
    '''
        Calculates rate of success in plotting ops
    '''
    return 0 if len(stat) == 0 else reduce(lambda acc, cur:
                                           acc+1 if cur else acc, stat, 0)/len(stat) * 100


def _parallelPlotting(chat: Chat, emojiData: List[int], sinkDirectory: str, extension: str) -> float:
    '''
        Implements process based paralleism using `ray` module,
        all plotting work done concurrently !!!
    '''
    _ids = [
        plotContributionInChatByUser.remote(
            chat,
            join(sinkDirectory,
                 'participationInChatByUser.{}'
                 .format(extension)),
            'Participation of Users in Chat [ {} - {} ]'
            .format(chat.startDate.strftime('%d %b, %Y'),
                    chat.endDate.strftime('%d %b, %Y'))
        ),
        *list(
            map(
                lambda cur:
                plotContributionOfUserByHour.remote(
                    cur.messages,
                    join(sinkDirectory,
                         'contributionInChatOf{}ByHour.{}'.format(
                             '_'.join(cur.name.split(' ')), extension)),
                    '{}\'s Contribution in Chat [ {} - {} ]'
                    .format(cur.name,
                            chat.startDate.strftime('%d %b, %Y'),
                            chat.endDate.strftime('%d %b, %Y'))
                ),
                chat.users
            )
        ),
        *list(
            map(
                lambda cur:
                plotActivityOfUserByMinute.remote(
                    cur.messages,
                    join(sinkDirectory, 'detailedActivityOf{}InChatByMinute.{}'
                         .format(
                             '_'.join(cur.name.split(' ')), extension)),
                    'Detailed Activity Of {} in Chat By Minute [ {} - {} ]'
                    .format(cur.name,
                            chat.startDate.strftime('%d %b, %Y'),
                            chat.endDate.strftime('%d %b, %Y'))
                ),
                chat.users
            )
        ),
        plotActivenessOfChatByDate.remote(
            classifyMessagesOfChatByDate(
                mergeMessagesFromUsersIntoSequence(chat)),
            join(sinkDirectory, 'activenessOfChatByDate.{}'
                 .format(extension)),
            'Daily Activeness Of a Chat [ {} - {} ]'
            .format(chat.startDate.strftime('%d %b, %Y'),
                    chat.endDate.strftime('%d %b, %Y'))
        ),
        plotConversationInitializerStat.remote(
            getConversationInitializers(chat),
            join(sinkDirectory,
                 'conversationInitializerStat.{}'.format(extension)),
            ('Conversation Initializers\' Statistics, using Mean Delay [ {} - {} ]'
             .format(chat.startDate.strftime('%d %b, %Y'),
                     chat.endDate.strftime('%d %b, %Y')),
             'Conversation Initializers\' Statistics, using Median Delay [ {} - {} ]'
             .format(chat.startDate.strftime('%d %b, %Y'),
                     chat.endDate.strftime('%d %b, %Y')))
        ),
        plotEmojiUsage.remote(
            findEmojiUsage(findEmojisInText(
                findNonASCIICharactersinText(chat),
                emojiData)),
            join(sinkDirectory,
                 'emojiUsage.{}'.format(extension)),
            'Top 7 Emoji(s) used in Chat [ {} - {} ]'
            .format(chat.startDate.strftime('%d %b, %Y'),
                    chat.endDate.strftime('%d %b, %Y'))
        ),
        *list(
            map(
                lambda cur:
                plotActivityHeatMap.remote(
                    cur.messages,
                    join(sinkDirectory, 'activityHeatMapOf{}InChat.{}'
                         .format(
                             '_'.join(cur.name.split(' ')), extension)),
                    'Activity HeatMap Of {} in Chat [ {} - {} ]'
                    .format(cur.name,
                            chat.startDate.strftime('%d %b, %Y'),
                            chat.endDate.strftime('%d %b, %Y'))
                ),
                chat.users
            )
        )
    ]

    # awaiting result of all plotting functions invoked
    return _calculatePercentageOfSuccess(ray.get(_ids))


def main():

    # extracts passed command line arguments
    def _getCMDArgs() -> Tuple[str, str]:
        return tuple(argv[1:len(argv)]) if len(argv) == 3 else (None, None)

    # prints usage of this script
    def _banner():
        print('\x1b[1;6;36;49m[+]chanalyze v0.3.7 - A simple WhatsApp Chat Analyzer\x1b[0m\n\n\t\x1b[3;30;47m$ chanalyze `path-to-exported-chat-file` `path-to-sink-directory`\x1b[0m\n\n[+]Author: Anjan Roy<anjanroy@yandex.com>\n[+]Source: https://github.com/itzmeanjan/chanalyze ( MIT Licensed )\n')

    startTime = time()
    endTime = startTime
    successRate = 0.0

    try:
        sourceFile, sinkDirectory = _getCMDArgs()
        # path to source file must be ending with `txt`, cause it's generally exported into a text file
        if not (sourceFile and sinkDirectory and sourceFile and sinkDirectory):
            _banner()
            raise Exception('Improper invokation !')

        directoryBuilder(sinkDirectory)
        if not (sourceFile.endswith('txt') and exists(sourceFile) and exists(sinkDirectory)):
            _banner()
            raise Exception('Invalid chat file !')

        _banner()
        # this instance will live throughout lifetime of this script
        chat = Chat.importFromText(sourceFile)
        emojiData = importFromFile() or exportToFile()

        if not emojiData:
            raise Exception('Unable to fetch emoji data')

        extension = _makeOutputChoice()
        if not extension:
            raise Exception('Invalid output format !')

        print('[*]Working ...')
        successRate = _parallelPlotting(
            chat, emojiData, sinkDirectory, extension)
        endTime = time()
    except KeyboardInterrupt:
        print('\n[!]Terminated')
        endTime = time()
    except Exception as e:
        print('[!]{}'.format(e))
        endTime = time()
    finally:
        print('[+]Success : {} % in {}s'
              .format(successRate,
                      endTime - startTime))
        ray.shutdown()
        return


if __name__ == '__main__':
    main()



#!/usr/bin/python3

from __future__ import annotations
from typing import List
from datetime import time


class MessagesSentInMinute(object):
    def __init__(self, hour: int, minute: int, count: int):
        self.hour = hour
        self.minute = minute
        self.count = count

    def incrementCount(self, by: int = 1):
        self.count += 1


class MessagesSentInADay(object):
    def __init__(self, records: List[MessagesSentInMinute]):
        self.records = records

    def findARecord(self, findIt: time) -> MessagesSentInMinute:
        return self.records[findIt.hour*60 + findIt.minute]


if __name__ == '__main__':
    print('[!]This module is expected to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from os import getenv
from os.path import (
    join,
    dirname,
    abspath
)
from requests import get as getContent
from re import compile as regCompile
from functools import reduce
from typing import List

'''
    Designed to extract out all supported
    emojis from Unicode.org ( 1311 in count, as per v12.0 )

    Takes a target url ( from where it'll fetch these data ),
    though a default value is provided, may be when next version
    of unicode releases, I'll update default value.

    Returns all supported emojis as List[int]
'''


def _getEmojiData(url: str = 'https://unicode.org/Public/emoji/12.0/emoji-data.txt') -> List[int]:
    '''
        There're certain cases where a code can be given
        in following form

        `xxxx..xxxx`

        For covering all values hiding in that
        range, I wrote following utility funtion,
        which will returns a List[int], holding
        all those values, which will eventually be
        appended to accumulated List[int]
    '''
    def __split__(e: str) -> List[int]:
        return [i for i in range(*[(int(j, base=16) + 1) if i == 1 else int(j, base=16)
                                   for i, j in enumerate(e.split('..'))])]
    try:
        # without custom headers remote was blocking request, thinking it's a bot
        resp = getContent(url, headers={
            'Accept': 'text/html,application/xhtml+xml,application/xml,application/json;q=0.9,*/*;q=0.8',
            'Accept-Language': 'en-US,en;q=0.5',
            'Connection': 'keep-alive',
            'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:60.0) Gecko/20100101 Firefox/71.0'
        })
        if not resp.ok:
            raise Exception('Bad Response from Remote')
        # regex to filter out lines which are having emojis listed
        regEmoji = regCompile(
            r'^[0-9A-F]{4,}([\.\.0-9A-F]+)?\s+;\s*Emoji\s+\#.+$')
        regHexCode = regCompile(r'^[0-9A-F]{4,}(\S+)?(?=\s+;)')
        return reduce(
            lambda acc, cur: acc +
            __split__(cur) if '..' in cur else acc + [int(cur, base=16)],
            map(lambda e: regHexCode.search(e).group(),
                filter(lambda e:
                       (not e.startswith('#')) and regEmoji.search(e),
                       resp.content.decode('utf-8').splitlines())), [])
    except Exception:
        return None


def exportToFile(sink: str = 'emoji.txt') -> List[int]:
    '''
        Utility function to generate a text file
        containing all supported emojis ( as per Unicode v12.0 )
        along with their corresponding numeric form ( integer value )

        Example :

        .
        .
        .
        11088,⭐
        11093,⭕
        12336,〰
        12349,〽
        .
        .
        .
    '''

    try:
        data = _getEmojiData()
        with open(abspath(join(dirname(__file__), '..', sink)), 'w') as fd:
            fd.writelines(
                map(lambda e: '{},{}\n'.format(e, chr(e)), data))
        return data
    except Exception:
        return None


def importFromFile(source: str = 'emoji.txt') -> List[int]:
    '''
        Given path to emoji file, imports unicode unsigned integer data,
        representing emojis, which can be used for generating plots
    '''

    try:
        with open(abspath(join(dirname(__file__), '..', source)),
                  'r') as fd:
            return list(map(lambda e: int(e.split(',')[0], base=10),
                            fd.readlines()))
    except Exception:
        return None


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from functools import reduce
from unicodedata import name as getNameOfUnicodeChar
from typing import List
from collections import Counter
from matplotlib.colors import cnames
from matplotlib import pyplot as plt
import ray

from .model.chat import Chat

'''
    Extracts all those non-ascii characters
    from Chat, which are to be filtered further,
    so that we can keep only those unicode characters,
    which are emojis.

    And understand emoji usage of this Chat
'''


def findNonASCIICharactersinText(chat: Chat) -> List[str]:
    return reduce(lambda acc, cur:
                  acc +
                  reduce(lambda accInner, curInner:
                         accInner +
                         reduce(lambda accInnerDeep, curInnerDeep:
                                (accInnerDeep + [curInnerDeep]) if ord(
                                    curInnerDeep) > 255 else accInnerDeep,
                                curInner.content, []),
                         cur.messages, []),
                  chat.users, [])


'''
    Filters out only those unicode characters,
    which are emojis, from a list of all 
    non-ascii characters used in this Chat,
    using a list of integers, which is nothing but
    all supported emojis numeric representation
'''


def findEmojisInText(data: List[str], emojiData: List[int]) -> List[str]:
    return reduce(lambda acc, cur:
                  acc + [cur] if ord(cur) in emojiData else acc,
                  data, [])


'''
    Returns a statistics of which emojis was used how many
    times in a certain Chat ( which is currently under consideration ),
    from a list of all used emojis in this Chat
'''


def findEmojiUsage(emojis: List[str]) -> Counter:
    return Counter(emojis)


@ray.remote
def plotEmojiUsage(emojis: Counter, targetPath: str, title: str, top: int = 7) -> bool:
    try:
        _tmpLabels = reduce(lambda acc, cur: acc + [cur] if len(acc) < (top+1) else acc, sorted(
            emojis, key=lambda e: emojis[e], reverse=True), [])
        total = sum([emojis[i] for i in emojis])
        _remaining = total - sum([emojis[i] for i in _tmpLabels])
        data = [emojis[i] for i in _tmpLabels] + \
            [_remaining] if len(_tmpLabels) < len(emojis) else []
        labels = ['{} ( {:7.4f} % )'.format(getNameOfUnicodeChar(i).title(), emojis[i]*100/total)
                  for i in _tmpLabels] + \
            ['Others ( {:7.4f} % )'.format(_remaining*100/total)] if len(
                _tmpLabels) < len(emojis) else []
        font = {
            'family': 'serif',
            'color': '#000000',
            'weight': 'normal',
            'size': 16
        }
        plt.figure(figsize=(24, 12), dpi=100)
        patches, _ = plt.pie(data)
        plt.legend(patches, labels, loc='best', fontsize='large')
        plt.title(title, fontdict=font)
        plt.axis('equal')
        plt.tight_layout()
        plt.savefig(targetPath, bbox_inches='tight',
                    pad_inches=.6, quality=95, dpi=100)  # exporting plotted PIE chart
        plt.close()  # closing this figure on which we just plotted a PIE chart
        return True
    except Exception:
        return False


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from typing import List

from .message import Message

'''
    Holds all messages sent by a certain user, participated in a chat
    ( for both private & group ), along with its name
'''


class User(object):
    def __init__(self, name: str, messages: List[Message]):
        self.name = name
        self.messages = messages

    def __str__(self):
        super().__str__()
        return '{} with {} messages'.format(self.name, len(self.messages))


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from datetime import date

'''
    Holds # of messages sent on a certain date,
    denoted by dd/mm/yyyy
'''


class MessagesSentOnDate(object):
    def __init__(self, currentDate: date, count: int):
        self.currentDate = currentDate
        self.count = count

    def incrementBy(self, by: int = 1):
        self.count += by


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from typing import Tuple


class TimePeriod(object):
    def __init__(self, init: int, end: int, step: int):
        self.init = init
        self.end = end
        self.step = step

    def __contains__(self, value: int) -> bool:
        return value >= self.init and value < self.end

    def __str__(self):
        super().__str__()

        def __secondToString__(second: int) -> str:
            def __convert__(dividend: int, divisor: int) -> Tuple[int, Tuple[int, int]]:
                if divisor == 1:
                    return dividend
                else:
                    return dividend//divisor, __convert__(dividend % divisor, divisor//60)
            return '{0:02}:{1[0]:02}:{1[1]:02}'.format(*__convert__(second, 3600))

        return '{} - {}'.format(__secondToString__(self.init), __secondToString__(self.end-1))


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



'''
A simple WhatsApp Chat Analyzer ( for both Private & Group chats ), made with <3

- Chanalyze aims to be a easy tool, for analyzing your Whatsapp Chat(s).
- No data sent to remote, everything stays in your machine.

[+]Author: Anjan Roy<anjanroy@yandex.com>
[+]Source: https://github.com/itzmeanjan/chanalyze ( MIT Licensed )
'''

__version__ = '0.3.7'



#!/usr/bin/python3

from __future__ import annotations
from os.path import abspath, dirname, exists
from os import mkdir
from functools import reduce
from typing import Dict, List, Tuple, Any
from collections import OrderedDict, Counter
from math import ceil, sqrt
from datetime import datetime, date, time
from matplotlib import pyplot as plt
from matplotlib.ticker import MultipleLocator, PercentFormatter, StrMethodFormatter, NullLocator, NullFormatter
from matplotlib.dates import HourLocator, DateFormatter, MinuteLocator, MonthLocator, DayLocator
import ray
import seaborn as sns

from .model.chat import Chat
from .model.message import Message, MessageIndex
from .model.timePeriod import TimePeriod
from .model.msgInMinute import MessagesSentInMinute, MessagesSentInADay
from .model.msgOnDate import MessagesSentOnDate
from .model.msgDiff import DifferenceBetweenMessages


def directoryBuilder(targetPath: str):
    '''
        Checks whether this directory already exists or not.
        if not, creates it
    '''
    dirName = abspath(targetPath)
    if not exists(dirName):
        mkdir(dirName)


def shadeContactName(name: str, percent: float = 50.0, where: str = 'f') -> str:
    '''
        Shades first half of characters of Contact by `*`, for sake of privacy
    '''
    return '*'*ceil(len(name)*percent/100)\
        + name[ceil(len(name)*percent/100):] if where.lower() == 'f' else name[:-ceil(len(name)*percent/100)]\
        + '*'*ceil(len(name)*percent/100)


@ray.remote
def plotContributionInChatByUser(chat: Chat, targetPath: str, title: str, top: int = 25) -> bool:
    '''
        How much a certain person contributed to a Chat 
        plotted into a Bar Chart ( shows contribution in terms of percentage )
    '''
    try:
        # considering only top 25 contributors in Chat
        y = reduce(lambda acc, cur: acc + [cur] if len(acc) < (top+1) else acc,
                   sorted([i.name for i in chat.users],
                          key=lambda e: len(chat.getUser(e).messages),
                          reverse=True),
                   [])
        # y = sorted([i.name for i in chat.users],
        #           key=lambda e: len(chat.getUser(e).messages), reverse=True)
        y_pos = range(len(y))
        x = [(len(chat.getUser(i).messages) / chat.messageCount) * 100 for i in y]
        # y = [shadeContactName(i, percent=75) for i in y]
        with plt.style.context('Solarize_Light2'):
            font = {
                'family': 'serif',
                'color': '#000000',
                'weight': 'normal',
                'size': 10
            }
            fig = plt.figure(figsize=(24, 12), dpi=100)
            plt.xlim((0, 100))
            fig.gca().xaxis.set_major_locator(MultipleLocator(10))
            fig.gca().xaxis.set_major_formatter(PercentFormatter())
            fig.gca().xaxis.set_minor_locator(MultipleLocator(1))
            fig.gca().yaxis.set_ticks(y_pos)
            fig.gca().yaxis.set_ticklabels(y)
            plt.barh(y_pos, x, align='center',
                     color='deepskyblue', lw=1.6)
            plt.xlabel('Percentage of Participation in Chat',
                       fontdict=font, labelpad=12)
            plt.title(title,
                      fontdict=font, pad=12)
            fig.tight_layout()
            fig.savefig(targetPath, bbox_inches='tight', pad_inches=.5)
            plt.close(fig=fig)
        return True
    except Exception:
        return False


@ray.remote
def plotContributionOfUserByHour(messages: List[Message], targetPath: str, title: str) -> bool:

    def __buildStatusHolder__(holder: Dict[str, int], current: Message) -> Dict[str, int]:
        tm = current.timeStamp.timetz()
        seconds = tm.hour*3600 + tm.minute*60
        found = str(
            reduce(lambda acc, cur: cur if seconds in cur else acc, ranges, None))
        holder.update({found: holder.get(found, 0)+1})
        return holder

    try:
        ranges = [TimePeriod(i, i+3600*3, 1)
                  for i in range(0, 86400, 3600*3)]
        splitted = reduce(__buildStatusHolder__, messages,
                          dict([(str(i), 0) for i in ranges]))
        _tmpLabels = sorted(splitted)
        data = [splitted[i] for i in _tmpLabels]
        splitted = OrderedDict([(i, splitted[i])
                                for i in _tmpLabels])
        _tmpLabels = [i for i in splitted]
        data = [splitted[i] for i in _tmpLabels]
        total = sum(data)
        labels = ['{} ( {:7.4f} % )'.format(i, splitted[i]*100/total)
                  for i in _tmpLabels]
        font = {
            'family': 'serif',
            'color': '#000000',
            'weight': 'normal',
            'size': 16
        }
        plt.figure(figsize=(24, 12), dpi=100)
        patches, _ = plt.pie(data)  # plotting pie chart
        plt.legend(patches, labels, loc='best', fontsize='medium')
        plt.title(title, fontdict=font)
        plt.axis('equal')
        plt.tight_layout()
        plt.savefig(targetPath, bbox_inches='tight',
                    pad_inches=.5)  # exporting plotted PIE chart
        plt.close()  # closing this figure on which we just plotted a PIE chart
        return True
    except Exception:
        return False


@ray.remote
def plotActivityOfUserByMinute(messages: List[Message], targetPath: str, title: str) -> bool:
    '''
        Plots a chart, showing at which minute of
        day ( there's 1440 minutes in a day )
        this participant is how much active ( in this chat ),
        over period of time, for which we've track record ( in exported chat )
    '''

    def _determineMajorLocatorSpacing(maxV: int) -> int:
        '''
            Determines how to place major & minor locators on both axes,
            so that we don't over-populate axes

            Max allowed number of major locators 20
        '''
        maxV = maxV if maxV % 10 == 0 else (maxV - maxV % 10 + 10)
        return round(sqrt(maxV))

    def __buildEachMinuteStatHolder__(holder: MessagesSentInADay, current: Message) -> MessagesSentInADay:
        '''
            This is to be called for each element
            present in `messages`, so that
            we can keep track of #-of messages
            sent in each minute
        '''
        holder.findARecord(current.timeStamp.timetz()).incrementCount()
        return holder

    def __splitIntoParts__(whole, partCount: int):
        '''
            Splits a collection of objects into
            equal parts of requested count i.e.
            returns a collection of ( sub ) collections
        '''
        if not whole:
            return [None] * partCount

        lengthOfEachPart = len(whole)//partCount
        return [whole[i*lengthOfEachPart:(i+1)*lengthOfEachPart]
                for i in range(partCount)]

    try:
        statByMinute = reduce(__buildEachMinuteStatHolder__, messages, MessagesSentInADay([
            MessagesSentInMinute(i, j, 0) for i in range(0, 24) for j in range(0, 60)]))
        tmpX = [datetime.combine(date(2000, 1, 1), time(i.hour, i.minute))
                for i in statByMinute.records]
        x1, x2, x3, x4 = __splitIntoParts__(tmpX, 4)
        if not (x1 and x2 and x3 and x4):
            return False

        y1, y2, y3, y4 = __splitIntoParts__(
            [statByMinute.findARecord(i.timetz()).count for i in tmpX], 4)
        if not (y1 and y2 and y3 and y4):
            return False

        # this will help us in setting max value on Y-axis
        # and no doubt min value is 0
        maxMsgCount = max([max([max([max(y1)] + y2)] + y3)] + y4) + 1
        locatorSpacingAlongY = _determineMajorLocatorSpacing(maxMsgCount)
        with plt.style.context('ggplot'):
            font = {
                'family': 'serif',
                'color': '#000000',
                'weight': 'normal',
                'size': 10
            }
            _, ((axes1, axes2), (axes3, axes4)) = plt.subplots(
                2, 2, figsize=(24, 12), dpi=100)  # creating 4 subplots in a Figure
            axes1.xaxis.set_major_locator(HourLocator())
            axes1.xaxis.set_major_formatter(DateFormatter('%I:%M %p'))
            axes1.xaxis.set_minor_locator(MinuteLocator())
            axes2.xaxis.set_major_locator(HourLocator())
            axes2.xaxis.set_major_formatter(DateFormatter('%I:%M %p'))
            axes2.xaxis.set_minor_locator(MinuteLocator())
            axes3.xaxis.set_major_locator(HourLocator())
            axes3.xaxis.set_major_formatter(DateFormatter('%I:%M %p'))
            axes3.xaxis.set_minor_locator(MinuteLocator())
            axes4.xaxis.set_major_locator(HourLocator())
            axes4.xaxis.set_major_formatter(DateFormatter('%I:%M %p'))
            axes4.xaxis.set_minor_locator(MinuteLocator())
            # setting formatting for Y-axis ( for all subplots )
            axes1.yaxis.set_major_locator(
                MultipleLocator(locatorSpacingAlongY))
            axes1.yaxis.set_major_formatter(StrMethodFormatter('{x}'))
            axes1.yaxis.set_minor_locator(NullLocator())
            axes2.yaxis.set_major_locator(
                MultipleLocator(locatorSpacingAlongY))
            axes2.yaxis.set_major_formatter(StrMethodFormatter('{x}'))
            axes2.yaxis.set_minor_locator(NullLocator())
            axes3.yaxis.set_major_locator(
                MultipleLocator(locatorSpacingAlongY))
            axes3.yaxis.set_major_formatter(StrMethodFormatter('{x}'))
            axes3.yaxis.set_minor_locator(NullLocator())
            axes4.yaxis.set_major_locator(
                MultipleLocator(locatorSpacingAlongY))
            axes4.yaxis.set_major_formatter(StrMethodFormatter('{x}'))
            axes4.yaxis.set_minor_locator(NullLocator())
            # setting limit of tickers along X axis ( time axis )
            # axes1.set_xlim(x1[0], x1[-1])
            # axes2.set_xlim(x2[0], x2[-1])
            # axes3.set_xlim(x3[0], x3[-1])
            # axes4.set_xlim(x4[0], x4[-1])
            # setting limit of ticker values along Y axis ( #-of messages sent in that minute )
            axes1.set_ylim(0, maxMsgCount)
            axes2.set_ylim(0, maxMsgCount)
            axes3.set_ylim(0, maxMsgCount)
            axes4.set_ylim(0, maxMsgCount)
            # plotting data
            axes1.plot(x1, y1, 'r-', lw=.5)
            axes2.plot(x2, y2, 'r-', lw=.5)
            axes3.plot(x3, y3, 'r-', lw=.5)
            axes4.plot(x4, y4, 'r-', lw=.5)
            axes1.set_xlabel('Time',
                             fontdict=font, labelpad=12)
            axes1.set_ylabel('#-of Messages Sent',
                             fontdict=font, labelpad=12)
            axes2.set_xlabel('Time',
                             fontdict=font, labelpad=12)
            axes2.set_ylabel('#-of Messages Sent',
                             fontdict=font, labelpad=12)
            axes3.set_xlabel('Time',
                             fontdict=font, labelpad=12)
            axes3.set_ylabel('#-of Messages Sent',
                             fontdict=font, labelpad=12)
            axes4.set_xlabel('Time',
                             fontdict=font, labelpad=12)
            axes4.set_ylabel('#-of Messages Sent',
                             fontdict=font, labelpad=12)
            axes1.set_title(title,
                            fontdict=font, pad=12)
            axes2.set_title(title,
                            fontdict=font, pad=12)
            axes3.set_title(title,
                            fontdict=font, pad=12)
            axes4.set_title(title,
                            fontdict=font, pad=12)
            plt.tight_layout()
            plt.savefig(targetPath, bbox_inches='tight',
                        pad_inches=.2, quality=95, dpi=100)  # exporting plotting into a file ( image )
            plt.close()  # this is required, closing drawing canvas
        return True
    except Exception:
        return False


def mergeMessagesFromUsersIntoSequence(chat: Chat) -> List[Message]:
    '''
        Takes a Chat object & returns a sequenced list of messages,
        as they appeared in original chat
    '''
    currentMsgIdx = 0
    nextMessageToBeVisitedForEachParticipant = [0]*len(chat.users)
    sequence = []
    for _ in range(chat.messageCount):
        for j, k in enumerate(chat.users):
            try:
                # there may be a case when we're trying to reach to some message index for a certain user `k`, which doesn't even exist
                # then we may require to handle one exception
                msg = k.messages[nextMessageToBeVisitedForEachParticipant[j]]
                if msg.index == currentMsgIdx:
                    sequence.append(msg)
                    currentMsgIdx += 1
                    nextMessageToBeVisitedForEachParticipant[j] += 1
                    break
            except Exception:
                continue
    return sequence


def classifyMessagesOfChatByDate(messages: List[Message]) -> List[MessagesSentOnDate]:
    '''
        Classifies a collection of messages
        by Date on which they were sent.

        So no doubt this may produce really large dataset in case
        of lengthy chats

        Returns a collection of MessagesSentOnDate objects,
        each of them holding # of messages trasmitted
        among chat participants on that certain date
    '''
    def __updateStat__(acc: List[MessagesSentOnDate], cur: Message) -> List[MessagesSentOnDate]:
        found = reduce(lambda accInner, curInner: curInner if curInner.currentDate ==
                       cur.timeStamp.date() else accInner, acc, None)
        if not found:
            acc.append(MessagesSentOnDate(cur.timeStamp.date(), 1))
        else:
            found.incrementBy()
        return acc
    return reduce(__updateStat__, messages, [])


@ray.remote
def plotActivenessOfChatByDate(messages: List[MessagesSentOnDate], targetPath: str, title: str) -> bool:
    '''
        Tries to depict how all participating users of a Chat
        contributed to traffic of that Chat by Day

        If we've a very long chat ( spreading across years ),
        then we'll simply accumulate it into a year ( 
        holding whole record into 365/366 days )

        It can be useful in understanding at which day of Year,
        people of this chat preferred to talk mostly
    '''
    def _determineMajorLocatorSpacing(data: List[int]) -> int:
        '''
            Determines how to place major & minor locators on both axes,
            so that we don't over-populate axes

            Max allowed number of major locators 20
        '''
        maxV = max(data)
        # finds out next round number
        maxV = maxV if maxV % 10 == 0 else (maxV - maxV % 10 + 10)
        return round(sqrt(maxV))

    def __accumulateData__(data: List[MessagesSentOnDate]) -> List[MessagesSentOnDate]:
        def __updateCount__(acc: List[MessagesSentOnDate], cur: MessagesSentOnDate) -> List[MessagesSentOnDate]:
            found = reduce(lambda accInner, curInner: curInner if cur.currentDate.day ==
                           curInner.currentDate.day and cur.currentDate.month == curInner.currentDate.month
                           else accInner, acc, None)
            if not found:
                acc.append(MessagesSentOnDate(date(
                    1996, cur.currentDate.month, cur.currentDate.day), cur.count))  # using a year, which was leap year, cause it won't throw one error when trying to define a date having 29th day in Feb
            else:
                found.incrementBy(cur.count)
            return acc
        return reduce(__updateCount__, data, [])
    try:
        messages = __accumulateData__(messages)
        x = [i.currentDate for i in messages]
        y = [i.count for i in messages]
        majorLocatorCount = _determineMajorLocatorSpacing(y)
        with plt.style.context('Solarize_Light2'):
            font = {
                'family': 'serif',
                'color': '#000000',
                'weight': 'normal',
                'size': 16
            }
            plt.figure(figsize=(24, 12), dpi=100)
            plt.gca().xaxis.set_major_locator(MonthLocator())
            plt.gca().xaxis.set_major_formatter(DateFormatter('%b'))
            plt.gca().xaxis.set_minor_locator(DayLocator())
            plt.gca().xaxis.set_minor_formatter(DateFormatter('%d'))
            plt.gca().yaxis.set_major_locator(MultipleLocator(majorLocatorCount))
            plt.gca().yaxis.set_major_formatter(StrMethodFormatter('{x}'))
            plt.gca().yaxis.set_minor_locator(MultipleLocator(majorLocatorCount / 2))
            plt.ylim(-5, max(y)+5)
            plt.plot(x, y, 'ro-', lw=.8)
            plt.xlabel('Time', fontdict=font, labelpad=14)
            plt.ylabel('#-of messages transferred', fontdict=font, labelpad=14)
            plt.title(title, fontdict=font, pad=14)
            plt.tight_layout()
            plt.savefig(targetPath, bbox_inches='tight',
                        pad_inches=.4, quality=95, dpi=100)  # exporting plotting into a file ( image )
            plt.close()
        return True
    except Exception:
        return False


def getConversationInitializers(chat: Chat) -> Tuple[Counter, Counter]:
    '''
        Get how many times which chat participant started
        a conversation ( applicable for both Private & Group chat )

        By a conversation in a Chat, I mean, when a collection of messages
        are/ were transmitted between these participants, after a certain delay
        & may be lasted for a while

        For finding that I calculated elapsed time between all messages
        of this Chat. Now I find unique delay values. From that meanDelay
        & medianDelay

        Now I filter out all those message senders
        who sent some message ( using message index )
        having delay value greater than or equal to meanDelay
        & medianDelay, and find frequency of them
    '''
    messages = mergeMessagesFromUsersIntoSequence(chat)
    diff = [DifferenceBetweenMessages(i, i+1, int((messages[i+1].timeStamp - j.timeStamp).total_seconds()))
            for i, j in enumerate(messages[:-1])]
    unique = sorted(reduce(lambda acc, cur: [
        cur] + acc if cur not in acc else acc, diff, []),
        key=lambda e: e.elapsedTime)
    if not unique:
        return (None, None)

    meanDelay = sum([i.elapsedTime for i in unique])//len(unique)
    medianDelay = unique[len(unique)//2].elapsedTime
    return (Counter(map(lambda e: chat.getUserByMessageId(e.msgTwo).name,
                        filter(lambda e: e.elapsedTime >= meanDelay, diff))),
            Counter(map(lambda e: chat.getUserByMessageId(e.msgTwo).name,
                        filter(lambda e: e.elapsedTime >= medianDelay, diff))))


@ray.remote
def plotConversationInitializerStat(data: Tuple[Counter, Counter], targetPath: str, title: Tuple[str, str]) -> bool:
    try:
        if not (data[0] and data[1]):
            return False

        y1, y2 = [i for i in data[0]], [i for i in data[1]]
        x1, x2 = [data[0][i] for i in y1], [data[1][i] for i in y2]
        # contact name/ number shading is temporarily disabled
        # y1, y2 = [shadeContactName(i, percent=75) for i in y1], [shadeContactName(i, percent=75) for i in y2]
        total1, total2 = sum(x1), sum(x2)
        if not (total1 and total2):
            return False

        x1, x2 = [i*100/total1 for i in x1], [i*100/total2 for i in x2]
        with plt.style.context('Solarize_Light2'):
            font = {
                'family': 'serif',
                'color': '#000000',
                'weight': 'normal',
                'size': 12
            }
            _, (axes1, axes2) = plt.subplots(1, 2, figsize=(24, 12), dpi=12)
            # handling X-axis ticks & labels for subplot 1 ( on left )
            axes1.set_xlim(0, 100)
            axes1.xaxis.set_major_locator(MultipleLocator(10))
            axes1.xaxis.set_major_formatter(PercentFormatter())
            axes1.xaxis.set_minor_locator(MultipleLocator(1))
            axes1.xaxis.set_minor_formatter(NullFormatter())
            # handling X-axis ticks & labels for subplot 2 ( on right )
            axes2.set_xlim(0, 100)
            axes2.xaxis.set_major_locator(MultipleLocator(10))
            axes2.xaxis.set_major_formatter(PercentFormatter())
            axes2.xaxis.set_minor_locator(MultipleLocator(1))
            axes2.xaxis.set_minor_formatter(NullFormatter())
            axes1.barh(y1, x1, align='center', color='deepskyblue', lw=1.5)
            axes2.barh(y2, x2, align='center', color='deepskyblue', lw=1.5)
            axes1.set_xlabel('Percentage of Conversations Started',
                             fontdict=font, labelpad=16)
            axes2.set_xlabel('Percentage of Conversations Started',
                             fontdict=font, labelpad=16)
            axes1.set_title(title[0], fontdict=font, pad=16)
            axes2.set_title(title[1], fontdict=font, pad=16)
            plt.tight_layout()
            plt.savefig(targetPath, bbox_inches='tight',
                        pad_inches=.4, quality=95, dpi=100)
            plt.close()  # don't miss this, it's required. Otherwise it might result into memory leaking
            # And no doubt too much memory will stay occupied
        return True
    except Exception:
        return False


def prepareHeatMapData(data: List[Message]) -> Tuple[List[Dict[str, int]], List[str]]:
    '''
        Given list of messages sent on a date, for a 
        certain user for chat under inspection, it'll create a
        2D array, holding number of messages sent by user, over weekdays

        Another list to be returned too, holding weekdays chronologically
    '''
    if not data:
        return None, None

    data = classifyMessagesOfChatByDate(data)

    if not data:
        return None, None

    heatMapData = [{} for i in range(7)]
    weekNumber = []

    for i in data:

        weekInfo = 'Week {} of {}'.format(
            int(i.currentDate.strftime('%U'), base=10) + 1,
            i.currentDate.strftime('%Y')
        )

        heatMapData[int(i.currentDate.strftime('%w'), base=10)].update({
            weekInfo: i.count
        })

        if weekInfo not in weekNumber:
            weekNumber.append(weekInfo)

    _plottableData = [[] for i in range(7)]

    for i, _week in enumerate(heatMapData):
        tmp = _plottableData[i]

        for k in weekNumber:
            tmp.append(_week.get(k, 0))

    return _plottableData, weekNumber


@ray.remote
def plotActivityHeatMap(data: List[Message], targetPath: str, title: str) -> bool:
    '''
        Plots heatmap of user activity across week days
    '''
    try:
        if not data:
            return False

        font = {
            'family': 'serif',
            'color': '#000000',
            'weight': 'normal',
            'size': 16
        }

        _data, _weekIndices = prepareHeatMapData(data)

        if not (_data and _weekIndices):
            return False

        fig = plt.figure(figsize=(24, 12), dpi=100)
        axes = sns.heatmap(
            _data,
            cmap='Blues',
            lw=.5)

        axes.set_xticklabels(_weekIndices, rotation=90)
        axes.set_yticklabels(['Sunday', 'Monday', 'Tuesday', 'Wednesday',
                              'Thursday', 'Friday', 'Saturday'], rotation=0)
        axes.set_title(title, fontdict=font, pad=14)
        axes.set_xlabel('Week of Year', fontdict=font, labelpad=14)
        axes.set_ylabel('Week Day', fontdict=font, labelpad=14)

        plt.tight_layout()
        plt.savefig(targetPath, bbox_inches='tight',
                    pad_inches=.4, quality=95, dpi=100)
        plt.close(fig)
        return True
    except Exception:
        return False


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['chanalyze', 'chanalyze.emoji_data', 'chanalyze.model']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib', 'requests', 'ray', 'seaborn']

extras_require = \
{'dev': ['flit']}

entry_points = \
{'console_scripts': ['chanalyze = chanalyze.app:main']}

setup(name='chanalyze',
      version='0.3.7',
      description='A simple WhatsApp Chat Analyzer ( for both Private & Group chats ), made with <3',
      author='Anjan Roy',
      author_email='anjanroy@yandex.com',
      url='https://itzmeanjan.github.io/chanalyze',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



#!/usr/bin/python3

from __future__ import annotations
from datetime import datetime
from math import ceil

'''
    Used for holding all messages for
    a certain chat ( for both group & private )
'''


class Message(object):
    def __init__(self, index: int, content: str, timeStamp: str):
        self.index = index
        self.content = content.rstrip()
        self._timeStamp = timeStamp

    '''
        Returns timestamp of a message as a datetime
        object ( parsed from string timestamp )
    '''
    @property
    def timeStamp(self) -> datetime:
        try:
            return datetime.strptime(self._timeStamp, r'%d/%m/%y, %I:%M %p')
        except ValueError:
            return datetime.strptime(self._timeStamp, r'%d/%m/%Y, %I:%M %p')

    def __str__(self):
        super().__str__()
        return '{} - {}***... - {}'.format(self.index, self.content[:ceil(len(self.content)/9)], str(self.timeStamp))


'''
    A temporary class, used for keeping track of message indices,
    while parsing Chat from text file
'''


class MessageIndex(object):
    def __init__(self, idx: int):
        self.index = idx

    def increment(self, by: int = 1):
        self.index += by


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)
